;;; compat.el --- Compatibility Library              -*- lexical-binding: t; -*-

;; Copyright (C) 2021 Free Software Foundation, Inc.

;; Author: Philip Kaludercic <philipk@posteo.net>
;; Version: 28.1.0.0
;; Keywords: lisp
;; Package-Requires: ((emacs "24.1"))

;; This program is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation, either version 3 of the License, or
;; (at your option) any later version.

;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <https://www.gnu.org/licenses/>.

;;; Commentary:

;; To allow for the usage of Emacs functions and macros that are
;; defined in newer versions of Emacs, compat.el provides definitions
;; that are installed if necessary.  These reimplementations of
;; functions and macros are at least subsets of the actual
;; implementations.
;;
;; Not every function provided in newer versions of Emacs is provided
;; here.  Some depend on new features from the core, others cannot be
;; implemented to a meaningful degree.  The main audience for this
;; library are not regular users, but package maintainers.  Therefore
;; commands and user options are usually not implemented here.

;;; Code:

(eval-and-compile
  (defvar compat-generate-definition nil
    "Always generate the compatibility function.
Otherwise the function is only generated when the required
version is greater than the current Emacs version."))

(eval-when-compile
  ;; The macros and utility functions defined in this block are not of
  ;; interest to the transitive users of compat.el, and thus can be
  ;; "compiled away" when the package is installed.

  (defmacro compat-ignore (&rest _) "Ignore all arguments" nil)
  
  (defmacro compat-defvar (name initval docstring &rest attr)
    "Declare compatibility variable NAME with initial value INITIAL.
The obligatory documentation string DOCSTRING must be given.  The
remaining arguments form a plist, modifying the behaviour of this
macro. These include the obligatory :version attribute,
specifying what version this variable was defined in (or t if it
should always be defined), :feature indicating if the variable
should be loaded after a feature, :override that with a non-nil
value always aliases the compatibility variable, :noalias does
the opposite, and :local that either makes the variable permanent
local with a value of `permanent' or just buffer local with any
non-nil value."
    (declare (debug (name form stringp [&rest keywordp sexp]))
	     (doc-string 3) (indent 2))
    (let ((version (plist-get attr :version))
	  (feature (plist-get attr :feature))
	  (override (plist-get attr :override))
	  (noalias (plist-get attr :noalias))
	  (localp (plist-get attr :local))
	  (constantp (plist-get attr :constant)))
      ;; The macros generates nothing if either the current Emacs
      ;; version is greater or equal to the specified version and
      ;; `compat-generate-definition' doesn't force that
      ;; compatibility code is generated.
      (when (or (eq version t)
		(not (version< version emacs-version))
		compat-generate-definition)
	;; The actual implementation of the compatibility function
	;; is always defined with a `compat--' prefix, assuming.
	(let ((realname (intern (format "compat--%S" name))))
	  `(progn
	     ;; In case the function is part of a library that has
	     ;; to be loaded, we will declare the signature here so
	     ;; to satisfy the byte compiler.
	     ,(let ((body `(;; Compatibility variable definition:
			    (,(if constantp 'defconst 'defvar)
			     ,realname ,initval
			     ;; Prepend compatibility notice to the actual
			     ;; documentation string.
			     ,(format
			       "[Compatibility variable for `%S', defined in Emacs %s]\n\n%s"
			       name version docstring))
			    ;; Make variable as local if necessary
			    ,(cond
			      ((eq localp 'permanent)
			       `(put ',realname 'permanent-local t))
			      (localp
			       `(make-variable-buffer-local ',realname)))
			    ;; Alias compatibility variable to actual
			    ;; variable if requested and allowed.
			    (,@(cond
				(override '(progn))
				(noalias '(compat-ignore))
				(`(unless (boundp ',name))))
			     (defvaralias ',name ',realname)))))
		(if feature
		    `(eval-after-load ',feature
		       (lambda () ,@body))
		  `(progn ,@body))))))))
  
  (defun compat-common-fdefine (type name arglist docstring rest)
    "Generate compatibility code for a function NAME.
TYPE is one of `defun', for functions and `defmacro' for macros, and `advice'
ARGLIST is passed on directly to the definition, and DOCSTRING is
prepended with a compatibility note.  REST contains the remaining
definition, that may begin with a property list of attributes."
    (let ((body rest))
      (while (keywordp (car body))
	(setq body (cddr body)))
      (let ((version (plist-get rest :version))
	    (min-version (plist-get rest :min-version))
	    (max-version (plist-get rest :min-version))
	    (feature (plist-get rest :feature))
	    (override (plist-get rest :override))
	    (lexical (plist-get rest :let)))
	;; The macros generates nothing if either the current Emacs
	;; version is greater or equal to the specified version and
	;; `compat-generate-definition' doesn't force that
	;; compatibility code is generated.
	(when (or (eq version t)
		  (not (version< version emacs-version))
		  (and min-version (version< min-version emacs-version))
		  (and max-version (version< emacs-version max-version))
		  compat-generate-definition)
	  ;; The actual implementation of the compatibility function
	  ;; is always defined with a `compat--' prefix, assuming.
	  (let* ((realname (intern (format "compat--%S" name)))
		 (body `(let ,lexical
			  ;; Compatibility functionality
			  ;; definition:
			  (,(cond
			     ((memq type '(func advice)) 'defun)
			     ((eq type 'macro) 'defmacro)
			     ((error "Unknown type")))
			   ,realname ,arglist
			   ;; Prepend compatibility notice to the actual
			   ;; documentation string.
			   ,(format
			     "[Compatibility %s for `%S', defined in Emacs %s]\n\n%s"
			     (cond
			      ((eq type 'func) "function")
			      ((eq type 'macro) "macro")
			      ((eq type 'advice) "advice")
			      ((error "Unknown type")))
			     name version docstring)
			   ;; Advice may use the implicit variable `oldfun', but
			   ;; to avoid triggering the byte compiler, we make
			   ;; sure the argument is used at least once.
			   ,(and (eq type 'advice) '(ignore oldfun))
			   ;; Insert the actual body now.
			   ,@body)
			  (,@(cond
			      (override '(progn))
			      ((version< emacs-version version)
			       '(compat-ignore))
			      (`(unless (fboundp ',name))))
			   ,(cond
			     ((memq type '(func macro))
			      ;; Functions and macros are installed by
			      ;; aliasing the name of the compatible
			      ;; function to the name of the compatibility
			      ;; function.
			      `(defalias ',name #',realname))
			     ((eq type 'advice)
			      ;; Advice is installed the usual way.
			      `(advice-add ',name :around #',realname)))))))
	    `(progn
	       ;; In case the function is part of a library that has
	       ;; to be loaded, we will declare the signature here so
	       ;; to satisfy the byte compiler.
	       (declare-function ,realname "compat" ',arglist)
	       ,(if feature
		    `(eval-after-load ',feature
		       (lambda () ,body))
		  body)))))))

  (defmacro compat-defun (name arglist docstring &rest rest)
    "Define NAME with arguments ARGLIST as a compatibility function.
The function must be documented in DOCSTRING.  REST may begin
with a plist, that is interpreted by the macro but not passed on
to the actual function.

The definition will only be loaded, if the version this function
was defined in, as indicated by the :version attribute, is
greater than the current Emacs version.

Given a :feature attribute, the definition will be wrapped in a
`eval-after-load' block, so that the user still has to
require the feature."
    (declare (debug (&define name (&rest symbolp)
			     stringp
			     [&rest keywordp sexp]
			     def-body))
	     (doc-string 3) (indent 2))
    (compat-common-fdefine 'func name arglist docstring rest))

  (defmacro compat-defmacro (name arglist docstring &rest rest)
    "Define NAME with arguments ARGLIST as a compatibility macro.
The macro must be documented in DOCSTRING.  REST may begin
with a plist, that is interpreted by this macro but not passed on
to the actual macro.

The definition will only be loaded, if the version this function
was defined in, as indicated by the :version attribute, is
greater than the current Emacs version.

Given a :feature attribute, the definition will be wrapped in a
`eval-after-load' block, so that the user still has to
require the feature."
    (declare (debug compat-defun) (doc-string 3) (indent 2))
    (compat-common-fdefine 'macro name arglist docstring rest))

  (defmacro compat-advise (name arglist docstring &rest rest)
    "Define NAME with arguments ARGLIST as a compatibility advice.
The advice function must be documented in DOCSTRING.  REST may
begin with a plist, that is interpreted by this macro but not
passed on to the actual advice function.  The advice wraps the
old definition, that is accessible via using the symbol `oldfun'.

The definition will only be loaded, if the version this function
was defined in, as indicated by the :version attribute, is
greater than the current Emacs version.

Given a :feature attribute, the definition will be wrapped in a
`eval-after-load' block, so that the user still has to
require the feature."
    (declare (debug compat-defun) (doc-string 3) (indent 2))
    (compat-common-fdefine 'advice name (cons 'oldfun arglist) docstring rest)))


;;; Loaded by default
;;;; Defined in Emacs 28

(compat-defun string-search (needle haystack &optional start-pos)
  "Search for the string NEEDLE in the strign HAYSTACK.

The return value is the position of the first occurrence of
NEEDLE in HAYSTACK, or nil if no match was found.

The optional START-POS argument says where to start searching in
HAYSTACK and defaults to zero (start at the beginning).
It must be between zero and the length of HAYSTACK, inclusive.

Case is always significant and text properties are ignored."
  :version "27.0"
  (let ((haystack (substring haystack start-pos)))
    (save-match-data
      (and (string-match (regexp-quote needle) haystack)
	   (match-beginning 0)))))

(compat-defun string-replace (fromstring tostring instring)
  "Replace FROMSTRING with TOSTRING in INSTRING each time it occurs."
  :version "28.1"
  (when (equal fromstring "")
    (signal 'wrong-length-argument '(0)))
  (save-match-data
    (replace-regexp-in-string
     (regexp-quote fromstring)
     tostring
     instring)))

(compat-defun length= (sequence length)
  "Returns non-nil if SEQUENCE has a length equal to LENGTH."
  :version "28.1"
  (cond
   ((listp sequence)
    (and (null (nthcdr length sequence))
	 (nthcdr (1- length) sequence)))
   ((arrayp sequence)
    (= (length sequence) length))
   ((error "Unsupported sequence type"))))

(compat-defun length< (sequence length)
  "Returns non-nil if SEQUENCE is shorter than LENGTH."
  :version "28.1"
  (cond
   ((listp sequence)
    (not (null (nthcdr length sequence))))
   ((arrayp sequence)
    (< (length sequence) length))
   ((error "Unsupported sequence type"))))

(compat-defun length> (sequence length)
  "Returns non-nil if SEQUENCE is longer than LENGTH."
  :version "28.1"
  (cond
   ((listp sequence)
    (nthcdr length sequence))
   ((arrayp sequence)
    (> (length sequence) length))
   ((error "Unsupported sequence type"))))

(compat-defun always (&rest _arguments)
  "Do nothing and return t.
This function accepts any number of ARGUMENTS, but ignores them.
Also see `ignore'."
  :version "28.1"
  t)

(compat-defun insert-into-buffer (buffer &optional start end)
  "Insert the contents of the current buffer into BUFFER.
If START/END, only insert that region from the current buffer.
Point in BUFFER will be placed after the inserted text."
  :version "28.1"
  (let ((current (current-buffer)))
    (with-current-buffer buffer
      (insert-buffer-substring current start end))))

(compat-defun replace-regexp-in-region (regexp replacement &optional start end)
  "Replace REGEXP with REPLACEMENT in the region from START to END.
The number of replaced occurrences are returned, or nil if REGEXP
doesn't exist in the region.

If START is nil, use the current point.  If END is nil, use `point-max'.

Comparisons and replacements are done with fixed case.

REPLACEMENT can use the following special elements:

  `\\&' in NEWTEXT means substitute original matched text.
  `\\N' means substitute what matched the Nth `\\(...\\)'.
       If Nth parens didn't match, substitute nothing.
  `\\\\' means insert one `\\'.
  `\\?' is treated literally."
  :version "28.1"
  (if start
      (when (< start (point-min))
        (error "Start before start of buffer"))
    (setq start (point)))
  (if end
      (when (> end (point-max))
        (error "End after end of buffer"))
    (setq end (point-max)))
  (save-excursion
    (let ((matches 0)
          (case-fold-search nil))
      (goto-char start)
      (while (re-search-forward regexp end t)
        (replace-match replacement t)
        (setq matches (1+ matches)))
      (and (not (zerop matches))
           matches))))

(compat-defun file-name-with-extension (filename extension)
  "Set the EXTENSION of a FILENAME.
The extension (in a file name) is the part that begins with the last \".\".

Trims a leading dot from the EXTENSION so that either \"foo\" or
\".foo\" can be given.

Errors if the FILENAME or EXTENSION are empty, or if the given
FILENAME has the format of a directory.

See also `file-name-sans-extension'."
  :version "28.1"
  (let ((extn (string-trim-left extension "[.]")))
    (cond
     ((string= filename "")
      (error "Empty filename"))
     ((string= extn "")
      (error "Malformed extension: %s" extension))
     ((directory-name-p filename)
      (error "Filename is a directory: %s" filename))
     (t
      (concat (file-name-sans-extension filename) "." extn)))))

(compat-defun file-name-concat (directory &rest components)
  "Append COMPONENTS to DIRECTORY and return the resulting string.
Elements in COMPONENTS must be a string or nil.
DIRECTORY or the non-final elements in COMPONENTS may or may not end
with a slash -- if they don’t end with a slash, a slash will be
inserted before contatenating."
  :version "28.1"
  (let ((seperator (eval-when-compile
		     (if (memq system-type '(ms-dos windows-nt cygwin))
			 "\\" "/"))))
    (mapconcat (lambda (part)
		 (replace-regexp-in-string
		  (concat seperator "+\\'") "" part))
	       (cons directory components)
	       seperator)))

(compat-defun directory-empty-p (dir)
  "Return t if DIR names an existing directory containing no other files.
Return nil if DIR does not name a directory, or if there was
trouble determining whether DIR is a directory or empty.

Symbolic links to directories count as directories.
See `file-symlink-p' to distinguish symlinks."
  :version "28.1"
  (and (file-directory-p dir)
       (null (directory-files dir nil directory-files-no-dot-files-regexp t 1))))

(compat-defun buffer-local-boundp (symbol buffer)
  "Return non-nil if SYMBOL is bound in BUFFER.
Also see `local-variable-p'."
  :version "28.1"
  (catch 'fail
    (condition-case nil
	(buffer-local-value symbol buffer)
      (void-variable nil (throw 'fail nil)))
    t))

(compat-defmacro with-existing-directory (&rest body)
  "Execute BODY with `default-directory' bound to an existing directory.
If `default-directory' is already an existing directory, it's not changed."
  :version "28.1"
  (declare (indent 0) (debug t))
  (let ((quit (gensym)))
    `(catch ',quit
       (dolist (dir (list default-directory
			  (expand-file-name "~/")
			  (getenv "TMPDIR")
			  "/tmp/"
			  ;; XXX: check if "/" works on non-POSIX
			  ;; system.
			  "/"))
	 (when (and dir (file-exists-p dir))
	   (throw ',quit (let ((default-directory dir))
			   ,@body)))))))

(compat-defun format-prompt (prompt default &rest format-args)
  "Format PROMPT with DEFAULT.
If FORMAT-ARGS is nil, PROMPT is used as a plain string.  If
FORMAT-ARGS is non-nil, PROMPT is used as a format control
string, and FORMAT-ARGS are the arguments to be substituted into
it.  See `format' for details.

If DEFAULT is a list, the first element is used as the default.
If not, the element is used as is.

If DEFAULT is nil or an empty string, no \"default value\" string
is included in the return value."
  :version "28.1"
  (concat
   (if (null format-args)
       prompt
     (apply #'format prompt format-args))
   (and default
        (or (not (stringp default))
            (length> default 0))
        (format " (default %s)"
                (if (consp default)
                    (car default)
                  default)))
   ": "))

(compat-defun garbage-collect-maybe (_factor)
  "Call ‘garbage-collect’ if enough allocation happened.
FACTOR determines what \"enough\" means here: If FACTOR is a
positive number N, it means to run GC if more than 1/Nth of the
allocations needed to trigger automatic allocation took place.
Therefore, as N gets higher, this is more likely to perform a GC.
Returns non-nil if GC happened, and nil otherwise."
  :version "28.1"
  ;; Do nothing
  nil)

;; TODO provide macro dlet
;; TODO provide macro named-let

(compat-advise unlock-buffer ()
  "Handle `file-error' conditions:

Handles file system errors by calling ‘display-warning’ and
continuing as if the error did not occur."
  :version "28.1"
  (condition-case error
      (funcall oldfun)
    (file-error
     (display-warning
      '(unlock-file)
      (message "%s, ignored" (error-message-string error))
      :warning))))

(compat-advise count-windows (&optional minibuf all-frames)
  "Handle optional argument ALL-FRAMES:

If ALL-FRAMES is non-nil, count the windows in all frames instead
just the selected frame."
  :version "28.1"
  (if all-frames
      (let ((sum 0))
	(dolist (frame (frame-list))
	  (with-selected-frame frame
	    (setq sum (+ (funcall oldfun minibuf)
			 sum))))
	sum)
    (funcall oldfun minibuf)))

(compat-advise string-width (string &optional from to)
  "Handle optional arguments FROM and TO:

Optional arguments FROM and TO specify the substring of STRING to
consider, and are interpreted as in `substring'."
  :version "28.1"
  (funcall oldfun (substring string from to)))

(compat-advise directory-files (directory &optional full match nosort count)
  "Handle additional optional argument COUNT:

If COUNT is non-nil and a natural number, the function will
 return COUNT number of file names (if so many are present)."
  :version "28.1"
  (let ((files (funcall oldfun directory full match nosort)))
    (when (natnump count)
      (setf (nthcdr count files) nil))
    files))

;;;; Defined in Emacs 27

(compat-advise setq-local (&rest pairs)
  "Handle multiple assignments."
  :version "27.1"
  (let (body)
    (while pairs
      (push `(set (make-local-variable ,(pop pairs))
		  (pop pairs))
	    body))
    (cons 'progn (nreverse body))))

(compat-defmacro ignore-error (condition &rest body)
  "Execute BODY; if the error CONDITION occurs, return nil.
Otherwise, return result of last form in BODY.

CONDITION can also be a list of error conditions."
  :version "28.1"
  (declare (debug t) (indent 1))
  `(condition-case nil (progn ,@body) (,condition nil)))

(compat-defun proper-list-p (object)
  "Return OBJECT's length if it is a proper list, nil otherwise.
A proper list is neither circular nor dotted (i.e., its last cdr
is nil)."
  :version "27.1"
  (catch 'improper
    (let ((length 0))
      (while object
	(unless (consp object)
	  (throw 'improper nil))
	(setq object (cdr object)
	      length (1+ length)))
      length)))

(compat-advise recenter (&optional arg redisplay)
  "Handle optional argument REDISPLAY."
  :version "27.1"
  (funcall oldfun arg)
  (when (and redisplay recenter-redisplay)
    (redisplay)))

(compat-defmacro dolist-with-progress-reporter (spec reporter-or-message &rest body)
  "Loop over a list and report progress in the echo area.
Evaluate BODY with VAR bound to each car from LIST, in turn.
Then evaluate RESULT to get return value, default nil.

REPORTER-OR-MESSAGE is a progress reporter object or a string.  In the latter
case, use this string to create a progress reporter.

At each iteration, print the reporter message followed by progress
percentage in the echo area.  After the loop is finished,
print the reporter message followed by the word \"done\".

\(fn (VAR LIST [RESULT]) REPORTER-OR-MESSAGE BODY...)"
  :version "27.1"
  (declare (indent 2) (debug ((symbolp form &optional form) form body)))
  (let ((prep (make-symbol "--dolist-progress-reporter--"))
        (count (make-symbol "--dolist-count--"))
        (list (make-symbol "--dolist-list--")))
    `(let ((,prep ,reporter-or-message)
           (,count 0)
           (,list ,(cadr spec)))
       (when (stringp ,prep)
         (setq ,prep (make-progress-reporter ,prep 0 (1- (length ,list)))))
       (dolist (,(car spec) ,list)
         ,@body
         (progress-reporter-update ,prep (setq ,count (1+ ,count))))
       (progress-reporter-done ,prep)
       (or ,@(cdr (cdr spec)) nil))))

(compat-defun flatten-tree (tree)
  "Return a \"flattened\" copy of TREE.
In other words, return a list of the non-nil terminal nodes, or
leaves, of the tree of cons cells rooted at TREE.  Leaves in the
returned list are in the same order as in TREE.

\(flatten-tree \\='(1 (2 . 3) nil (4 5 (6)) 7))
=> (1 2 3 4 5 6 7)"
  :version "27.1"
  (let (elems)
    (while (consp tree)
      (let ((elem (pop tree)))
        (while (consp elem)
          (push (cdr elem) tree)
          (setq elem (car elem)))
        (if elem (push elem elems))))
    (if tree (push tree elems))
    (nreverse elems)))

(compat-defun xor (cond1 cond2)
  "Return the boolean exclusive-or of COND1 and COND2.
If only one of the arguments is non-nil, return it; otherwise
return nil."
  :version "27.1"
  (declare (pure t) (side-effect-free error-free))
  (cond ((not cond1) cond2)
        ((not cond2) cond1)))

(compat-defun string-distance (string1 string2 &optional bytecompare)
  "Return Levenshtein distance between STRING1 and STRING2.
The distance is the number of deletions, insertions, and substitutions
required to transform STRING1 into STRING2.
If BYTECOMPARE is nil or omitted, compute distance in terms of characters.
If BYTECOMPARE is non-nil, compute distance in terms of bytes.
Letter-case is significant, but text properties are ignored."
  :version "27.1"
  ;; https://en.wikipedia.org/wiki/Levenshtein_distance
  (let ((s1 (if bytecompare
		(encode-coding-string string1 'raw-text)
	      string1))
	(s2 (if bytecompare
		(encode-coding-string string2 'raw-text)
	      string2)))
    (let* ((len1 (length s1))
	   (len2 (length s2))
	   (column (make-vector len1 0)))
      (dotimes (y len1)
	(setf (aref column y) y))
      (dotimes (x (1- len2))
	(setf (aref column 0) (1+ x))
	(let ((lastdiag x) olddiag)
	  (dotimes (y (1- len1))
	    (setf olddiag (aref column (1+ y))
		  (aref column (1+ y))
		  (min (+ (if (= (aref s1 y) (aref s2 x)) 0 1)
			  lastdiag)
		       (1+ (aref column (1+ y)))
		       (1+ (aref column y)))
		  lastdiag olddiag))))
      (aref column (1- len1)))))

(compat-defvar regexp-unmatchable "\\`a\\`"
  "Standard regexp guaranteed not to match any string at all."
  :version "27.1"
  :constant t)

(compat-advise regexp-opt (strings &optional paren)
  "Handle an empty list of strings."
  :version "27.1"
  (if (null strings)
      (funcall oldfun (list "\\`a\\`") paren)
    (funcall oldfun strings paren)))

;; TODO provide advice for file-size-human-readable
;; TODO provide advice for format-spec
;; TODO provide advice for directory-files-recursively
;; TODO provide advice for define-globalized-minor-mode

;;;; Defined in Emacs 26

(compat-advise assoc (key alist &optional testfn)
  "Handle TESTFN manually."
  :version "26.1"
  (if testfn
      (catch 'found
	(dolist (ent alist)
	  (when (funcall testfn (car ent) key)
	    (throw 'found ent))))
    (funcall oldfun key alist)))

(compat-advise alist-get (key alist &optional default remove testfn)
  "Handle TESTFN manually."
  :version "26.1"
  :min-version "25.1"			;first defined in 25.1
  :max-version "25.3"			;last version without testfn
  (if testfn
      (catch 'found
	(dolist (ent alist)
	  (when (funcall testfn (car ent) key)
	    (throw 'found (cdr ent))))
	default)
    (funcall oldfun key alist default remove testfn)))

(compat-defun string-trim-left (string &optional regexp)
  "Trim STRING of leading string matching REGEXP.

REGEXP defaults to \"[ \\t\\n\\r]+\"."
  :version "26.1"
  (if (string-match (concat "\\`\\(?:" (or regexp "[ \t\n\r]+") "\\)") string)
      (substring string (match-end 0))
    string))

(compat-defun string-trim-right (string &optional regexp)
  "Trim STRING of trailing string matching REGEXP.

REGEXP defaults to  \"[ \\t\\n\\r]+\"."
  :version "26.1"
  (let ((i (string-match-p
	    (concat "\\(?:" (or regexp "[ \t\n\r]+") "\\)\\'")
            string)))
    (if i (substring string 0 i) string)))

(compat-defun string-trim (string &optional trim-left trim-right)
  "Trim STRING of leading with and trailing matching TRIM-LEFT and TRIM-RIGHT.

TRIM-LEFT and TRIM-RIGHT default to \"[ \\t\\n\\r]+\"."
  :version "26.1"
  (string-trim-left (string-trim-right string trim-right) trim-left))


(compat-defun mapcan (func sequence)
  "Apply FUNC to each element of SEQUENCE.
Concatenate the results by altering them (using `nconc').
SEQUENCE may be a list, a vector, a boolean vector, or a string."
  :version "26.1"
  (let (result)
    (mapc
     (lambda (elem)
       (setq result (nconc result (funcall func elem))))
     sequence)
    result))

(compat-defun caaar (x)
  "Return the `car' of the `car' of the `car' of X."
  :version "26.1"
  (car (car (car x))))

(compat-defun caadr (x)
  "Return the `car' of the `car' of the `cdr' of X."
  :version "26.1"
  (car (car (cdr x))))

(compat-defun cadar (x)
  "Return the `car' of the `cdr' of the `car' of X."
  :version "26.1"
  (car (cdr (car x))))

(compat-defun caddr (x)
  "Return the `car' of the `cdr' of the `cdr' of X."
  :version "26.1"
  (car (cdr (cdr x))))

(compat-defun cdaar (x)
  "Return the `cdr' of the `car' of the `car' of X."
  :version "26.1"
  (cdr (car (car x))))

(compat-defun cdadr (x)
  "Return the `cdr' of the `car' of the `cdr' of X."
  :version "26.1"
  (cdr (car (cdr x))))

(compat-defun cddar (x)
  "Return the `cdr' of the `cdr' of the `car' of X."
  :version "26.1"
  (cdr (cdr (car x))))

(compat-defun cdddr (x)
  "Return the `cdr' of the `cdr' of the `cdr' of X."
  :version "26.1"
  (cdr (cdr (cdr x))))

(compat-defun caaaar (x)
  "Return the `car' of the `car' of the `car' of the `car' of X."
  :version "26.1"
  (car (car (car (car x)))))

(compat-defun caaadr (x)
  "Return the `car' of the `car' of the `car' of the `cdr' of X."
  :version "26.1"
  (car (car (car (cdr x)))))

(compat-defun caadar (x)
  "Return the `car' of the `car' of the `cdr' of the `car' of X."
  :version "26.1"
  (car (car (cdr (car x)))))

(compat-defun caaddr (x)
  "Return the `car' of the `car' of the `cdr' of the `cdr' of X."
  :version "26.1"
  (car (car (cdr (cdr x)))))

(compat-defun cadaar (x)
  "Return the `car' of the `cdr' of the `car' of the `car' of X."
  :version "26.1"
  (car (cdr (car (car x)))))

(compat-defun cadadr (x)
  "Return the `car' of the `cdr' of the `car' of the `cdr' of X."
  :version "26.1"
  (car (cdr (car (cdr x)))))

(compat-defun caddar (x)
  "Return the `car' of the `cdr' of the `cdr' of the `car' of X."
  :version "26.1"
  (car (cdr (cdr (car x)))))

(compat-defun cadddr (x)
  "Return the `car' of the `cdr' of the `cdr' of the `cdr' of X."
  :version "26.1"
  (car (cdr (cdr (cdr x)))))

(compat-defun cdaaar (x)
  "Return the `cdr' of the `car' of the `car' of the `car' of X."
  :version "26.1"
  (cdr (car (car (car x)))))

(compat-defun cdaadr (x)
  "Return the `cdr' of the `car' of the `car' of the `cdr' of X."
  :version "26.1"
  (cdr (car (car (cdr x)))))

(compat-defun cdadar (x)
  "Return the `cdr' of the `car' of the `cdr' of the `car' of X."
  :version "26.1"
  (cdr (car (cdr (car x)))))

(compat-defun cdaddr (x)
  "Return the `cdr' of the `car' of the `cdr' of the `cdr' of X."
  :version "26.1"
  (cdr (car (cdr (cdr x)))))

(compat-defun cddaar (x)
  "Return the `cdr' of the `cdr' of the `car' of the `car' of X."
  :version "26.1"
  (cdr (cdr (car (car x)))))

(compat-defun cddadr (x)
  "Return the `cdr' of the `cdr' of the `car' of the `cdr' of X."
  :version "26.1"
  (cdr (cdr (car (cdr x)))))

(compat-defun cdddar (x)
  "Return the `cdr' of the `cdr' of the `cdr' of the `car' of X."
  :version "26.1"
  (cdr (cdr (cdr (car x)))))

(compat-defun cddddr (x)
  "Return the `cdr' of the `cdr' of the `cdr' of the `cdr' of X."
  :version "26.1"
  (cdr (cdr (cdr (cdr x)))))

(compat-defun gensym (&optional prefix)
  "Return a new uninterned symbol.
The name is made by appending `gensym-counter' to PREFIX.
PREFIX is a string, and defaults to \"g\"."
  :version "26.1"
  :let ((counter 0))
  (let ((num (prog1 counter (setq counter (1+ counter)))))
    (make-symbol (format "%s%d" (or prefix "g") num))))

(compat-defun make-nearby-temp-file (prefix &optional dir-flag suffix)
  "Create a temporary file as close as possible to `default-directory'.
If PREFIX is a relative file name, and `default-directory' is a
remote file name or located on a mounted file systems, the
temporary file is created in the directory returned by the
function `temporary-file-directory'.  Otherwise, the function
`make-temp-file' is used.  PREFIX, DIR-FLAG and SUFFIX have the
same meaning as in `make-temp-file'."
  :version "26.1"
  (let ((handler (find-file-name-handler
                  default-directory 'make-nearby-temp-file)))
    (if (and handler (not (file-name-absolute-p default-directory)))
	(funcall handler 'make-nearby-temp-file prefix dir-flag suffix)
      (let ((temporary-file-directory (temporary-file-directory)))
        (make-temp-file prefix dir-flag suffix)))))

(compat-defun temporary-file-directory ()
  "The directory for writing temporary files.
In case of a remote `default-directory', this is a directory for
temporary files on that remote host.  If such a directory does
not exist, or `default-directory' ought to be located on a
mounted file system (see `mounted-file-systems'), the function
returns `default-directory'.
For a non-remote and non-mounted `default-directory', the value of
the variable `temporary-file-directory' is returned."
  :version "26.1"
  (let ((handler (find-file-name-handler
                  default-directory 'temporary-file-directory)))
    (if handler
	(funcall handler 'temporary-file-directory)
      (if (string-match mounted-file-systems default-directory)
          default-directory
        temporary-file-directory))))

(compat-advise line-number-at-pos (&optional position absolute)
  "Handle optional argument ABSOLUTE:

If the buffer is narrowed, the return value by default counts the lines
from the beginning of the accessible portion of the buffer.  But if the
second optional argument ABSOLUTE is non-nil, the value counts the lines
from the absolute start of the buffer, disregarding the narrowing."
  :version "26.1"
  (if absolute
      (save-restriction
	(widen)
	(funcall oldfun position))
    (funcall oldfun position)))

;;;; Defined in Emacs 25

(compat-defun string-greaterp (string1 string2)
  "Return non-nil if STRING1 is greater than STRING2 in lexicographic order.
Case is significant.
Symbols are also allowed; their print names are used instead."
  :version "25.1"
  (string-lessp string2 string1))

(compat-advise alist-get (key alist &optional default remove)
  "Handle TESTFN manually."
  :version "25.1"
  :max-version "25.3"
  (catch 'found
    (dolist (ent alist)
      (when (eq (car ent) key)
	(throw 'found (cdr ent))))
    default))

(compat-defmacro with-file-modes (modes &rest body)
  "Execute BODY with default file permissions temporarily set to MODES.
MODES is as for `set-default-file-modes'."
  :version "25.1"
  (declare (indent 1) (debug t))
  (let ((umask (make-symbol "umask")))
    `(let ((,umask (default-file-modes)))
       (unwind-protect
           (progn
             (set-default-file-modes ,modes)
             ,@body)
         (set-default-file-modes ,umask)))))

(compat-advise sort (seq predicate)
  "Handle SEQ of type vector."
  :version "25.1"
  (cond
   ((listp seq)
    (funcall oldfun seq predicate))
   ((vectorp seq)
    (let ((cseq (sort (append seq nil) predicate)))
      (dotimes (i (length cseq))
	(setf (aref seq i) (nth i cseq)))
      (apply #'vector cseq)))
   ((signal 'wrong-type-argument 'list-or-vector-p))))

(compat-defun format-message (string &rest objects)
  "Format a string out of a format-string and arguments.
The first argument is a format control string.
The other arguments are substituted into it to make the result, a string.

This implementation is equivalent to `format'."
  :version "25.1"
  (apply #'format string objects))

;; TODO advise read-buffer to handle 4th argument

;;;; Defined in Emacs 24.4

(compat-advise = (= number-or-marker &rest numbers-or-markers)
  "Handle multiple arguments."
  :version "24.4"
  (catch 'fail
    (while numbers-or-markers
      (unless (funcall oldfun number-or-marker (pop numbers-or-markers))
	(throw 'fail nil)))
    t))

(compat-advise < (number-or-marker &rest numbers-or-markers)
  "Handle multiple arguments."
  :version "24.4"
  (catch 'fail
    (while numbers-or-markers
      (unless (funcall oldfun number-or-marker (pop numbers-or-markers))
	(throw 'fail nil)))
    t))

(compat-advise > (number-or-marker &rest numbers-or-markers)
  "Handle multiple arguments."
  :version "24.4"
  (catch 'fail
    (while numbers-or-markers
      (unless (funcall oldfun number-or-marker (pop numbers-or-markers))
	(throw 'fail nil)))
    t))

(compat-advise <= (number-or-marker &rest numbers-or-markers)
  "Handle multiple arguments."
  :version "24.4"
  (catch 'fail
    (while numbers-or-markers
      (unless (funcall oldfun number-or-marker (pop numbers-or-markers))
	(throw 'fail nil)))
    t))

(compat-advise >= (number-or-marker &rest numbers-or-markers)
  "Handle multiple arguments."
  :version "24.4"
  (catch 'fail
    (while numbers-or-markers
      (unless (funcall oldfun number-or-marker (pop numbers-or-markers))
	(throw 'fail nil)))
    t))

(compat-defmacro with-eval-after-load (file &rest body)
  "Execute BODY after FILE is loaded.
FILE is normally a feature name, but it can also be a file name,
in case that file does not provide any feature.  See `eval-after-load'
for more details about the different forms of FILE and their semantics."
  :version "24.4"
  (declare (indent 1) (debug (form def-body)))
  `(eval-after-load ,file (lambda () ,@body)))

(compat-defun special-form-p (object)
  "Non-nil if and only if OBJECT is a special form."
  :version "24.4"
  (if (and (symbolp object) (fboundp object))
      (setq object (indirect-function object)))
  (and (subrp object) (eq (cdr (subr-arity object)) 'unevalled)))

(compat-defun macrop (object)
  "Non-nil if and only if OBJECT is a macro."
  :version "24.4"
  (let ((def (indirect-function object)))
    (when (consp def)
      (or (eq 'macro (car def))
          (and (autoloadp def) (memq (nth 4 def) '(macro t)))))))

(compat-defun string-suffix-p (suffix string  &optional ignore-case)
  "Return non-nil if SUFFIX is a suffix of STRING.
If IGNORE-CASE is non-nil, the comparison is done without paying
attention to case differences."
  :version "24.4"
  (let ((start-pos (- (length string) (length suffix))))
    (and (>= start-pos 0)
         (eq t (compare-strings suffix nil nil
                                string start-pos nil ignore-case)))))

(compat-advise split-string (string &optional separators omit-nulls trim)
  "Handle optional argument TRIM."
  :version "24.4"
  (let ((token (funcall oldfun string separators omit-nulls))
	(trimmed (if trim
		     (mapcar
		      (lambda (token)
			(when (string-match (concat "\\`" trim) token)
			  (setq token (substring token (match-end 0))))
			(when (string-match (concat trim "\\'") token)
			  (setq token (substring token 0 (match-beginning 0))))
			token)
		      token)
		   token)))
    (if omit-nulls (delete "" trimmed) trimmed)))

(compat-defun delete-consecutive-dups (list &optional circular)
  "Destructively remove `equal' consecutive duplicates from LIST.
First and last elements are considered consecutive if CIRCULAR is
non-nil."
  :version "24.4"
  (let ((tail list) last)
    (while (cdr tail)
      (if (equal (car tail) (cadr tail))
	  (setcdr tail (cddr tail))
	(setq last tail
	      tail (cdr tail))))
    (if (and circular
	     last
	     (equal (car tail) (car list)))
	(setcdr last nil)))
  list)

;;;;; Defined in Emacs 24.3

(compat-defmacro defvar-local (var val &optional docstring)
  "Define VAR as a buffer-local variable with default value VAL.
Like `defvar' but additionally marks the variable as being automatically
buffer-local wherever it is set."
  :version "24.3"
  (declare (debug defvar) (doc-string 3))
  ;; Can't use backquote here, it's too early in the bootstrap.
  (list 'progn (list 'defvar var val docstring)
        (list 'make-variable-buffer-local (list 'quote var))))

;;;;; Defined in Emacs 24.2

(compat-defun autoloadp (object)
  "Non-nil if OBJECT is an autoload."
  :version "24.2"
  (eq 'autoload (car-safe object)))

(compat-defun buffer-narrowed-p ()
  "Return non-nil if the current buffer is narrowed."
  :version "24.2"
  (/= (- (point-max) (point-min)) (buffer-size)))

(compat-defun file-name-base (filename)
  "Return non-nil if the current buffer is narrowed."
  :version "24.2"
  (file-name-sans-extension
   (file-name-nondirectory (or filename (buffer-file-name)))))

(compat-defun posnp (obj)
  "Return non-nil if OBJ appears to be a valid position.
A `posn' object is returned from functions such as `event-start'.
If OBJ is a valid `posn' object, but specifies a frame rather
than a window, return nil."
  :version "24.2"
  ;; FIXME: Correct the behavior of this function so that all valid
  ;; `posn' objects are recognized, after updating other code that
  ;; depends on its present behavior.
  (and (windowp (car-safe obj))
       (atom (car-safe (setq obj (cdr obj))))                ;AREA-OR-POS.
       (integerp (car-safe (car-safe (setq obj (cdr obj))))) ;XOFFSET.
       (integerp (car-safe (cdr obj)))))


;;; Part of subr-x
;;;; Defined in Emacs 28

(compat-defun string-clean-whitespace (string)
  "Clean up whitespace in STRING.
All sequences of whitespaces in STRING are collapsed into a
single space character, and leading/trailing whitespace is
removed."
  :version "28.1"
  :feature subr-x
  (let ((blank "[[:blank:]\r\n]+"))
    (string-trim (replace-regexp-in-string blank " " string t t)
                 blank blank)))

(compat-defun string-fill (string length)
  "Clean up whitespace in STRING.
All sequences of whitespaces in STRING are collapsed into a
single space character, and leading/trailing whitespace is
removed."
  :version "28.1"
  :feature subr-x
  (with-temp-buffer
    (insert string)
    (goto-char (point-min))
    (let ((fill-column length)
          (adaptive-fill-mode nil))
      (fill-region (point-min) (point-max)))
    (buffer-string)))

(compat-defun string-lines (string &optional omit-nulls)
  "Split STRING into a list of lines.
If OMIT-NULLS, empty lines will be removed from the results."
  :version "28.1"
  :feature subr-x
  (split-string string "\n" omit-nulls))

(compat-defun string-pad (string length &optional padding start)
  "Pad STRING to LENGTH using PADDING.
If PADDING is nil, the space character is used.  If not nil, it
should be a character.

If STRING is longer than the absolute value of LENGTH, no padding
is done.

If START is nil (or not present), the padding is done to the end
of the string, and if non-nil, padding is done to the start of
the string."
  :version "28.1"
  :feature subr-x
  (unless (natnump length)
    (signal 'wrong-type-argument (list 'natnump length)))
  (let ((pad-length (- length (length string))))
    (if (< pad-length 0)
        string
      (concat (and start
                   (make-string pad-length (or padding ?\s)))
              string
              (and (not start)
                   (make-string pad-length (or padding ?\s)))))))

(compat-defun string-chop-newline (string)
  "Remove the final newline (if any) from STRING."
  :version "28.1"
  :feature subr-x
  (if (= (aref string (1- (length string))) ?\n)
      (substring string 0 (- (length string) (length "\n")))
    string))

;;;; Defined in Emacs 25

(compat-defmacro if-let* (varlist then &rest else)
  "Bind variables according to VARLIST and evaluate THEN or ELSE.
This is like `if-let' but doesn't handle a VARLIST of the form
\(SYMBOL SOMETHING) specially."
  :version "25.1"
  :feature subr-x
  (declare (indent 2)
           (debug ((&rest [&or symbolp (symbolp form) (form)])
                   body)))
  (if varlist
      `(let* ,(setq varlist (internal--build-bindings varlist))
         (if ,(caar (last varlist))
             ,then
           ,@else))
    `(let* () ,then)))

(compat-defmacro when-let* (varlist &rest body)
  "Bind variables according to VARLIST and conditionally evaluate BODY.
This is like `when-let' but doesn't handle a VARLIST of the form
\(SYMBOL SOMETHING) specially."
  :version "25.1"
  :feature subr-x
  (declare (indent 1) (debug if-let*))
  (list 'if-let* varlist (macroexp-progn body)))

(compat-defmacro and-let* (varlist &rest body)
  "Bind variables according to VARLIST and conditionally evaluate BODY.
Like `when-let*', except if BODY is empty and all the bindings
are non-nil, then the result is non-nil."
  :version "25.1"
  :feature subr-x
  (declare (indent 1) (debug if-let*))
  (let (res)
    (if varlist
        `(let* ,(setq varlist (internal--build-bindings varlist))
           (when ,(setq res (caar (last varlist)))
             ,@(or body `(,res))))
      `(let* () ,@(or body '(t))))))

(compat-defmacro if-let (spec then &rest else)
  "Bind variables according to SPEC and evaluate THEN or ELSE.
Evaluate each binding in turn, as in `let*', stopping if a
binding value is nil.  If all are non-nil return the value of
THEN, otherwise the last form in ELSE.

Each element of SPEC is a list (SYMBOL VALUEFORM) that binds
SYMBOL to the value of VALUEFORM.  An element can additionally be
of the form (VALUEFORM), which is evaluated and checked for nil;
i.e. SYMBOL can be omitted if only the test result is of
interest.  It can also be of the form SYMBOL, then the binding of
SYMBOL is checked for nil.

As a special case, interprets a SPEC of the form \(SYMBOL SOMETHING)
like \((SYMBOL SOMETHING)).  This exists for backward compatibility
with an old syntax that accepted only one binding."
  :version "25.1"
  :feature subr-x
  (declare (indent 2)
           (debug ([&or (symbolp form)
                        (&rest [&or symbolp (symbolp form) (form)])]
                   body)))
  (when (and (<= (length spec) 2)
             (not (listp (car spec))))
    ;; Adjust the single binding case
    (setq spec (list spec)))
  
  (list 'if-let* spec then (macroexp-progn else)))

(compat-defmacro when-let (spec &rest body)
  "Bind variables according to SPEC and conditionally evaluate BODY.
Evaluate each binding in turn, stopping if a binding value is nil.
If all are non-nil, return the value of the last form in BODY.

The variable list SPEC is the same as in `if-let'."
  :version "25.1"
  :feature subr-x
  (declare (indent 1) (debug if-let))
  (list 'if-let spec (macroexp-progn body)))

(compat-defmacro thread-first (&rest forms)
  "Thread FORMS elements as the first argument of their successor.
Example:
    (thread-first
      5
      (+ 20)
      (/ 25)
      -
      (+ 40))
Is equivalent to:
    (+ (- (/ (+ 5 20) 25)) 40)
Note how the single `-' got converted into a list before
threading."
  :version "25.1"
  :feature subr-x
  (declare (indent 1)
           (debug (form &rest [&or symbolp (sexp &rest form)])))
  (let ((body (car forms)))
    (dolist (form (cdr forms))
      (when (symbolp form)
	(setq form (list form)))
      (setq body (append (list (car form))
			 (list body)
			 (cdr form))))
    body))

(compat-defmacro thread-last (&rest forms)
  "Thread FORMS elements as the last argument of their successor.
Example:
    (thread-last
      5
      (+ 20)
      (/ 25)
      -
      (+ 40))
Is equivalent to:
    (+ 40 (- (/ 25 (+ 20 5))))
Note how the single `-' got converted into a list before
threading."
  :version "25.1"
  :feature subr-x
  (declare (indent 1) (debug thread-first))
  (let ((body (car forms)))
    (dolist (form (cdr forms))
      (when (symbolp form)
	(setq form (list form)))
      (setq body (append form (list body))))
    body))


;;; Part of thingatpt
;;;; Defined in Emacs 28

(compat-defun thing-at-mouse (event thing &optional no-properties)
  "Return the THING at mouse click.
Like `thing-at-point', but tries to use the event
where the mouse button is clicked to find a thing nearby."
  :version "28.1"
  :feature thingatpt
  (save-excursion
    (mouse-set-point event)
    (thing-at-point thing no-properties)))


;;; Part of macroexp
;;;; Defined in Emacs 28

(compat-defun macroexp-file-name ()
  "Return the name of the file from which the code comes.
Returns nil when we do not know.
A non-nil result is expected to be reliable when called from a macro in order
to find the file in which the macro's call was found, and it should be
reliable as well when used at the top-level of a file.
Other uses risk returning non-nil value that point to the wrong file."
  :version "28.1"
  :feature macroexp
  (let ((file (car (last current-load-list))))
    (or (if (stringp file) file)
        (bound-and-true-p byte-compile-current-file))))

;;;; Defined in Emacs 25

(compat-defun macroexpand-1 (form &optional environment)
  "Perform (at most) one step of macro expansion."
  :version "25.1"
  :feature macroexp
  (cond
   ((consp form)
    (let* ((head (car form))
           (env-expander (assq head environment)))
      (if env-expander
          (if (cdr env-expander)
              (apply (cdr env-expander) (cdr form))
            form)
        (if (not (and (symbolp head) (fboundp head)))
            form
          (let ((def (autoload-do-load (symbol-function head) head 'macro)))
            (cond
             ;; Follow alias, but only for macros, otherwise we may end up
             ;; skipping an important compiler-macro (e.g. cl--block-wrapper).
             ((and (symbolp def) (macrop def)) (cons def (cdr form)))
             ((not (consp def)) form)
             (t
              (if (eq 'macro (car def))
                  (apply (cdr def) (cdr form))
                form))))))))
   (t form)))


;;; Part of package
;;;; Defined in Emacs 27

(declare-function lm-header "lisp-mnt")

(compat-defun package-get-version ()
  "Return the version number of the package in which this is used.
Assumes it is used from an Elisp file placed inside the top-level directory
of an installed ELPA package.
The return value is a string (or nil in case we can’t find it)."
  :version "27.1"
  ;; In a sense, this is a lie, but it does just what we want: precompute
  ;; the version at compile time and hardcodes it into the .elc file!
  (declare (pure t))
  ;; Hack alert!
  (let ((file
	 (or (and (boundp 'byte-compile-current-file) byte-compile-current-file)
	     load-file-name
	     buffer-file-name)))
    (cond
     ((null file) nil)
     ;; Packages are normally installed into directories named "<pkg>-<vers>",
     ;; so get the version number from there.
     ((string-match "/[^/]+-\\([0-9]\\(?:[0-9.]\\|pre\\|beta\\|alpha\\|snapshot\\)+\\)/[^/]+\\'" file)
      (match-string 1 file))
     ;; For packages run straight from the an elpa.git clone, there's no
     ;; "-<vers>" in the directory name, so we have to fetch the version
     ;; the hard way.
     ((let* ((pkgdir (file-name-directory file))
             (pkgname (file-name-nondirectory (directory-file-name pkgdir)))
             (mainfile (expand-file-name (concat pkgname ".el") pkgdir)))
        (when (file-readable-p mainfile)
          (require 'lisp-mnt)
          (with-temp-buffer
            (insert-file-contents mainfile)
            (or (lm-header "package-version")
                (lm-header "version")))))))))


;;; Etcetera

;; Is this necessary?
;;
;; ;;;; Suppress errors triggered by requiring non-existent libraries
;; (compat-advise require (feature &optional filename noerror)
;;   "Avoid throwing an error if library has compatibility code."
;;   :version t
;;   :override t
;;   (condition-case err
;;       (funcall oldfun feature filename noerror)
;;     (file-missing
;;      (unless (assq feature after-load-alist)
;;        (signal (car err) (cdr err))))))

;; Is this excusable?
;;
;; ;;;; Ensure NonGNU ELPA is in `package-archives'
;; (eval-after-load "package"
;;   (lambda ()
;;     (unless (or (rassoc "https://elpa.nongnu.org/nongnu/" package-archives)
;; 		(rassoc "http://elpa.nongnu.org/nongnu/" package-archives)
;; 		(rassoc "https://elpa.nongnu.org/nongnu-devel/" package-archives)
;; 		(rassoc "http://elpa.nongnu.org/nongnu-devel/" package-archives))
;;       (add-to-list 'package-archives '("nongnu" . "https://elpa.nongnu.org/nongnu/")))))

(provide 'compat)
;;; compat.el ends here
